import datetime
import re
from typing import Optional, Any, List, Union

from dateutil.parser import parse as dateutil_parse
from dateutil.tz import tzlocal, tzutc, UTC

# RFC3339: Date and Time on the Internet: Timestamps
#
# A string format for dates. A subset/profile of the ISO 8601 standard.
# examples:
#    2020-03-04T23:33:45.1234+01:00
#    2020-03-04T23:33:45.1234Z
#    2020-03-04T23:33:45Z
#
# See https://www.ietf.org/rfc/rfc3339.txt

RFC3339_PARSE_PATTERN_NOZULU = re.compile(
    r'([0-9][0-9][0-9][0-9])-([0-9][0-9])-([0-9][0-9])T([0-9][0-9]):([0-9][0-9]):([0-9][0-9])(.([0-9]+))?(([+-])([0-9][0-9]):00)')

RFC3339_PARSE_PATTERN_ZULU = re.compile(
    r'([0-9][0-9][0-9][0-9])-([0-9][0-9])-([0-9][0-9])T([0-9][0-9]):([0-9][0-9]):([0-9][0-9])(.([0-9]+))?Z')


def _fast_parse_rfc3339(text: str) -> Optional[datetime.datetime]:
    if (text.endswith('Z')):
        match = RFC3339_PARSE_PATTERN_ZULU.match(text)
        tzinfo = UTC
    else:
        match = RFC3339_PARSE_PATTERN_NOZULU.match(text)
        if match:
            assert match.group(10) in ('-', '+')
            sign = -1 if match.group(10) == '-' else 1
            tzhours = int(match.group(11))
            tzinfo = datetime.timezone(sign * datetime.timedelta(hours=tzhours, minutes=0))
        else:
            tzinfo = None
    if match:
        if match.group(7):
            msec_str = match.group(8)
            if len(msec_str) > 6:
                msec_str = msec_str[0:6]
            else:
                msec_str += '0' * (6 - len(msec_str))
            microsecond = int(msec_str)
        else:
            microsecond = 0
        return datetime.datetime(year=int(match.group(1)), month=int(match.group(2)), day=int(match.group(3)),
                          hour=int(match.group(4)), minute=int(match.group(5)), second=int(match.group(6)),
                          microsecond=microsecond,
                          tzinfo=tzinfo)
    else:
        return None


def parse_rfc3339(text: Optional[str], *, none_if_empty_tz=False) -> Optional[datetime.datetime]:
    """
    Parse an RFC3339 string representation of a time into a timezone aware datetime object
    Returns None if the provided string is None

    :param text: A string with an RFC3339 representation of a time
    :param none_if_empty_tz: If True, return None if the timezone in the string is empty. default: raise an error
    :return: A timezone aware datetime object, or None if the provided string is None
    """

    if text is None:
        return None
    # this is a lot faster than dateutil.parser.parse,
    # but might fail for some edge cases, or non-RFC3999 ISO8601 dates
    res = _fast_parse_rfc3339(text)
    if res is None:
        # This can handle more than just RFC3339 (full ISO 8601), but that's OK
        res = dateutil_parse(text)
    if res is not None and res.tzinfo is None:
        # enforce timezone aware (non-naive) datetime
        if none_if_empty_tz:
            return None
        else:
            raise ValueError('parsed date has no timezone: {!r}'.format(text))
    return res


def parse_rfc3339_no_none(text: str) -> datetime.datetime:
    """
    Parse an RFC3339 string representation of a time into a timezone aware datetime object
    Raises ValueError if the given string is None

    :param text: A string with an RFC3339 representation of a time
    :return: A timezone aware datetime object
    """
    res = parse_rfc3339(text)
    if res is None:
        raise ValueError('date parse error: {!r}'.format(text))
    return res


def dump_rfc3339(dt: Optional[datetime.datetime], zulu=True, no_milliseconds=True) -> str:
    """
    Convert the datetime object to an RFC3339 string representation of the time
    :param dt: The datetime to convert to string
    :param zulu: Force the timezone to "Zulu" (Z or UTC timezone)
    :param no_milliseconds: No sub second precision. This rounds the returned time down to the closest second.
    :return: An RFC3339 string representation of the provided datetime object, or None if the provided datetime is None.
    """
    if dt is None:
        return None
    assert isinstance(dt, datetime.datetime), 'not datetime.datetime but {}'.format(type(dt).__name__)
    if dt.tzinfo is None:
        raise ValueError('Naive datetimes are not supported. '
                         '(Because they can only cause trouble. Always use a timezone!)')
    assert dt.tzinfo is not None  # enforce timezone aware (non-naive) datetime
    if zulu:
        dt = dt.astimezone(tzutc())
    # else:
    #     dt = dt.astimezone(tzlocal())
    if no_milliseconds and dt.microsecond != 0:
        dt = dt.replace(microsecond=0)
    assert dt.tzinfo is not None  # enforce timezone aware (non-naive) datetime
    res = dt.isoformat()
    if res.endswith('+00:00'):
        res = res[:-6] + 'Z'
    return res


def datetime_now(zulu=False, no_milliseconds=True) -> datetime.datetime:
    """
    Get a timezone aware "now" time.

    :param zulu: force timezone to be "zulu" (Z or UTC timezone)
    :param no_milliseconds: No sub second precision. This rounds the returned time down to the closest second.
    :return: a timezone aware datetime object representing "now"
    """
    res = datetime.datetime.now(datetime.timezone.utc)
    if not zulu:
        res = res.astimezone(tzlocal())
    if no_milliseconds:
        res = res.replace(microsecond=0)
    assert res.tzinfo is not None  # enforce timezone aware (non-naive) datetime
    return res



import collections
import dataclasses
import json
from collections import Mapping
from datetime import datetime
from enum import Enum
from typing import Callable, Optional, Any, Dict, Type, Union, List

from dataclass_dict_convert.rfc3339 import dump_rfc3339, parse_rfc3339


class DataclassConvertError(Exception):
    def __init__(self, message):
        self.message = message


class UnknownFieldError(Exception):
    def __init__(self, message):
        self.message = message


@dataclasses.dataclass
class _DataclassDictConvertFieldMetaData:
    field_name: str
    dict_field_name: str
    to_dict_convertor: Callable[[Any], Any]
    from_dict_convertor: Callable[[Any], Any]


def _remove_none_recursive(val: Any) -> Any:
    if val is None:
        return None
    if isinstance(val, str) or isinstance(val, int) or isinstance(val, float) or isinstance(val, bool):
        return val
    if isinstance(val, list):
        # return [remove_none_recursive(v) for v in val if v is not None]
        return [_remove_none_recursive(v) for v in val]  # List is allowed to include None!
    if isinstance(val, dict):
        return {k: _remove_none_recursive(v) for k, v in val.items() if v is not None}
    raise ValueError('remove_none_recursive does not know how to handle type "{}"'.format(type(val)))


def _is_optional(a_type: type):
    if not (hasattr(a_type, '__origin__') and \
        a_type.__origin__ is Union and \
        len(a_type.__args__) == 2):
        return False
    t1 = a_type.__args__[0]
    t2 = a_type.__args__[1]
    for t in (t1,t2):
        if hasattr(t, '__origin__'):
            continue
        if isinstance(None, t):
            return True
    return False


def _get_optional_type(a_type: type):
    return a_type.__args__[0] if a_type.__args__[0] else a_type.__args__[1]


def _is_list(a_type: type):
    return hasattr(a_type, '__origin__') and a_type.__origin__ is list


def _is_dict(a_type: type):
    return hasattr(a_type, '__origin__') and a_type.__origin__ is dict


def _is_enum(a_type: type):
    if hasattr(a_type, '__origin__'):
        return False
    return issubclass(a_type, Enum)


def _is_json_primitive(a_type: type):
    return a_type is int or a_type is bool or a_type is float or a_type is str


def _find_convertor(
        field_name, field_type,
        default_datetime_convertor: Optional[Callable[[datetime], Any]],
        custom_dict_convertors: Dict[str, Callable[[Any], Any]],
        direct_fields: List[str],
        is_from: bool) -> Optional[Callable[[Any], Any]]:
    try:
        if field_name in custom_dict_convertors:
            return custom_dict_convertors[field_name]
        if _is_json_primitive(field_type) or field_name in direct_fields:
            return lambda p: p
        if field_type is datetime:
            return default_datetime_convertor
        if _is_optional(field_type):
            sub_convertor = _find_convertor(
                field_name, _get_optional_type(field_type),
                default_datetime_convertor, custom_dict_convertors, direct_fields,
                is_from)
            if sub_convertor is None:
                return None

            def _opt_convert_helper(orig_val):
                if orig_val is None:
                    return None
                else:
                    return sub_convertor(orig_val)
            return _opt_convert_helper
        if _is_list(field_type):
            sub_convertor = _find_convertor(
                field_name, field_type.__args__[0],
                default_datetime_convertor, custom_dict_convertors, direct_fields,
                is_from)
            if sub_convertor is None:
                return None

            def _opt_convert_helper(orig_val):
                return [sub_convertor(orig_el) for orig_el in orig_val]
            return _opt_convert_helper
        if _is_dict(field_type):
            if is_from:
                def dict_handler(d):
                    if not isinstance(d, Mapping):
                        raise DataclassConvertError(
                            "field {} of expected type dict is a {} instead".format(field_name, type(d)))
                    return dict(d)
                return dict_handler
            else:
                return lambda d: d
        if _is_enum(field_type):
            if is_from:
                return field_type.__members__.get
            else:
                return lambda enum_value: enum_value.name if enum_value is not None else None
        if dataclasses.is_dataclass(field_type):
            # assume dataclass has to_dict and from_dict
            if not hasattr(field_type, 'from_dict'):
                raise DataclassConvertError("Subclass '{}' of field {} does not have from_dict()".format(field_type, field_name))
            if not hasattr(field_type, 'to_dict'):
                raise DataclassConvertError("Subclass '{}' of field {} does not have to_dict()".format(field_type, field_name))
            if is_from:
                return lambda value: field_type.from_dict(value)
            else:
                return lambda value: value.to_dict()
    except:
        raise DataclassConvertError('Error while searching convertor for field "{}" of type "{}" is_from={}'
                                    .format(field_name, field_type, is_from))
    return None


def create_wrap_in_list_from_convertor(type_in_list: Type) -> Callable[[Any], List]:
    def wrap(val: str):
        val_type = type(val)
        if val_type is list:
            return val
        if val_type is type_in_list:
            return [val]
        raise TypeError('Unexpected type {} (expected {type_in_list} or List[{type_in_list}])'
                        .format(val_type, type_in_list=type_in_list))

    return wrap


def create_dict_of_dataclasses_from_convertor(type_of_dict_value, field_name_for_debug: Optional[str] = None) \
        -> Callable[[Dict[str, Any]], Dict]:
    def wrap(the_dict: dict):
        assert type(the_dict) is dict
        if not the_dict:
            return the_dict
        try:
            return {key: type_of_dict_value.from_dict(val) for key, val in the_dict.items()}
        except AttributeError:
            firstval = next(iter(the_dict.values()))
            raise DataclassConvertError(
                'Field {!r} has type {!r} which has no from_dict()'
                    .format(field_name_for_debug if field_name_for_debug else 'unknown', type(firstval)))

    return wrap


def create_dict_of_dataclasses_to_convertor(field_name_for_debug: Optional[str] = None) \
        -> Callable[[Dict[str, Any]], Dict]:
    def wrap(the_dict: dict):
        assert type(the_dict) is dict
        if not the_dict:
            return the_dict
        try:
            return {key: val.to_dict() for key, val in the_dict.items()}
        except AttributeError:
            _, firstval = next(iter(the_dict.values()))
            raise DataclassConvertError(
                'Field {} has type {} which has no to_dict()'
                    .format(field_name_for_debug if field_name_for_debug else 'unknown', type(firstval)))

    return wrap


@dataclasses.dataclass(frozen=True)
class _DataclassDictConvertMetaData:
    on_unknown_field: Callable[[str], None]
    metadata_by_fields: Dict[str, _DataclassDictConvertFieldMetaData]
    metadata_by_dict_fields: Dict[str, _DataclassDictConvertFieldMetaData]


def _wrap_dataclass_dict_convert(
        cls,
        dict_letter_case: Callable[[str], str],
        on_unknown_field: Callable[[str], None],
        direct_fields: List[str],
        default_to_datetime_convertor: Callable[[Any], datetime],
        default_from_datetime_convertor: Optional[Callable[[datetime], Any]],
        custom_to_dict_convertors: Dict[str, Callable[[Any], Any]],
        custom_from_dict_convertors: Dict[str, Callable[[Any], Any]]):
    metadata_by_fields = {}
    metadata_by_dict_fields = {}

    all_fields = dataclasses.fields(cls)

    for field in all_fields:
        meta = _DataclassDictConvertFieldMetaData(
            field.name,
            dict_letter_case(field.name),
            _find_convertor(field.name, field.type,
                            default_to_datetime_convertor, custom_to_dict_convertors, direct_fields,
                            is_from=False),
            _find_convertor(field.name, field.type,
                            default_from_datetime_convertor, custom_from_dict_convertors, direct_fields,
                            is_from=True)
        )
        if meta.to_dict_convertor is None:
            raise DataclassConvertError("""Unhandled "to" type '{}' of field '{}'""".format(field.type, field.name))
        if meta.from_dict_convertor is None:
            raise DataclassConvertError("""Unhandled "from" type '{}' of field '{}'""".format(field.type, field.name))
        metadata_by_fields[field.name] = meta
        metadata_by_dict_fields[meta.dict_field_name] = meta

    meta = _DataclassDictConvertMetaData(
        on_unknown_field, metadata_by_fields, metadata_by_dict_fields
    )
    # cls._dataclass_dict_convert_metadata = meta

    def _from_dict(cls2, d: dict, *, on_unknown_field_override: Optional[Callable[[str], None]] = None):
        if not isinstance(d, collections.Mapping):
            raise ValueError('from_dict(d) (possibly nested) where d is {} instead of dict: {!r}'.format(type(d), d))
        assert cls is cls2  # minor sanity check
        init_args = {}
        for key, value in d.items():
            field_meta = meta.metadata_by_dict_fields.get(key, None)
            if not field_meta:
                on_unknown_field_override(key) if on_unknown_field_override else meta.on_unknown_field(key)
            assert key == field_meta.dict_field_name
            init_args[field_meta.field_name] = field_meta.from_dict_convertor(value)
        return cls2(**init_args)

    def _to_dict(self, *, remove_none=False):
        res = {}
        for key, field_meta in meta.metadata_by_fields.items():
            assert key == field_meta.field_name
            value = getattr(self, field_meta.field_name)
            used_val = field_meta.to_dict_convertor(value)
            if used_val is not None or not remove_none:
                res[field_meta.dict_field_name] = _remove_none_recursive(used_val) if remove_none else used_val
        return res

    def _from_json(cls2, json_in: str, *, on_unknown_field_override: Optional[Callable[[str], None]] = None):
        assert cls is cls2  # minor sanity check
        return _from_dict(cls, json.loads(json_in), on_unknown_field_override=on_unknown_field_override)

    def _to_json(self) -> str:
        return json.dumps(self.to_dict())

    def _from_dict_list(cls2, l: list, *, on_unknown_field_override: Optional[Callable[[str], None]] = None):
        assert cls is cls2  # minor sanity check
        return [_from_dict(cls, d, on_unknown_field_override=on_unknown_field_override) for d in l]

    def _to_dict_list(inst_list: list):
        return [inst.to_dict() for inst in inst_list]

    cls.from_dict = classmethod(_from_dict)
    cls.to_dict = _to_dict
    cls.from_json = classmethod(_from_json)
    cls.to_json = _to_json
    cls.from_dict_list = classmethod(_from_dict_list)
    cls.to_dict_list = staticmethod(_to_dict_list)
    return cls


def dataclass_dict_convert(
        _cls=None, *,
        dict_letter_case: Optional[Callable[[str], str]]=None,
        on_unknown_field: Optional[Callable[[str], None]]=None,
        direct_fields: Optional[List[str]]=None,
        default_to_datetime_convertor: Optional[Callable[[Any], datetime]]=None,
        default_from_datetime_convertor: Optional[Callable[[datetime], Any]]=None,
        custom_to_dict_convertors: Optional[Dict[str, Callable[[Any], Any]]]=None,
        custom_from_dict_convertors: Optional[Dict[str, Callable[[Any], Any]]]=None):
    """
    This has complex logic to allow decorating with both:
      @dataclass_dict_convert
      @dataclass_dict_convert()

    dataclass_dict_convert takes parameters:

    example:
    @dataclass_dict_convert
    @dataclass(frozen=True)
    class DemoSub:
        int nr: int

    @dataclass_dict_convert(
         dict_letter_case=camelcase,
         on_undefined=None,  # default: raise exception
         default_to_datetime_convertor=parse_rfc3339,
         default_from_datetime_convertor=dump_rfc3339,
         custom_to_dict_convertors={
         },
         custom_from_dict_convertors={
             'str_lst': str_or_str_list_from_dict
         },
    )
    @dataclass(frozen=True)
    class Demo:
       nr: int
       text: str
       sometime: datetime
       opt_val: Optional[float]
       str_lst: List[str]
       sub: DemoSub
       list_sub: List[DemoSub]
       opt_sub: Optional[DemoSub]


    """
    def default_on_unknown(fieldname: str):
        raise UnknownFieldError('Unknown field encountered: {}'.format(fieldname))

    def wrap(cls):
        return _wrap_dataclass_dict_convert(
            cls,
            dict_letter_case if dict_letter_case else lambda s: s,
            on_unknown_field if on_unknown_field else default_on_unknown,
            direct_fields if direct_fields else [],
            default_to_datetime_convertor if default_to_datetime_convertor else dump_rfc3339,
            default_from_datetime_convertor if default_to_datetime_convertor else parse_rfc3339,
            custom_to_dict_convertors if custom_to_dict_convertors else {},
            custom_from_dict_convertors if custom_from_dict_convertors else {},)

    if _cls is None:
        return wrap
    return wrap(_cls)



from .rfc3339 import parse_rfc3339, dump_rfc3339, datetime_now
# rfc3339 also has parse_rfc3339_no_none

from .convert import dataclass_dict_convert, \
    create_wrap_in_list_from_convertor, \
    create_dict_of_dataclasses_from_convertor, create_dict_of_dataclasses_to_convertor, \
    DataclassConvertError, UnknownFieldError

from .dataclass_utils import dataclass_auto_type_check, dataclass_multiline_repr, dataclass_copy_method
# There's more in dataclass_utils, but these are the most useful

__all__ = [
    parse_rfc3339, dump_rfc3339, datetime_now, dataclass_dict_convert,
    create_wrap_in_list_from_convertor,
    create_dict_of_dataclasses_from_convertor, create_dict_of_dataclasses_to_convertor,
    DataclassConvertError, UnknownFieldError,
    dataclass_auto_type_check, dataclass_multiline_repr, dataclass_copy_method
]



import _thread
import copy
import dataclasses
import datetime
from typing import Optional, Any, List, Union

#
# dataclasses utils that have nothing to do with coverting to/from json/dict
#

def _field_type_name(field_type) -> str:
    try:
        return field_type.__name__
    except AttributeError:
        return '{!r}'.format(field_type)


def dataclass_field_simple_type_check(obj: Any, field_name: str, required_type: type) -> None:
    """
    use dataclass_field_auto_type_check instead!
    :param obj:
    :param field_name:
    :param required_type:
    :return:
    """
    field_val = getattr(obj, field_name)
    if not isinstance(field_val, required_type):
        raise TypeError("{}.{} must be {}, not {}".
                        format(type(obj).__name__, field_name, required_type.__name__, type(field_val).__name__))


def dataclass_field_type_check(obj: Any, field_name: str) -> None:
    field = next(field for field in dataclasses.fields(obj) if field.name == field_name)
    field_val = getattr(obj, field_name)
    if not isinstance(field_val, field.type):
        raise TypeError("{}.{} must be {}, not {}".
                        format(type(obj).__name__, field_name, _field_type_name(field.type), type(field_val).__name__))


def dataclass_field_datetime_check(obj: Any, field_name: str) -> None:
    field = next(field for field in dataclasses.fields(obj) if field.name == field_name)
    field_val = getattr(obj, field_name)
    assert field.type == datetime.datetime
    if not isinstance(field_val, datetime.datetime):
        raise TypeError("{}.{} must be datetime.datetime, not {}".
                        format(type(obj).__name__, field_name, type(field_val).__name__))
    if field_val.tzinfo is None:
        raise TypeError("{}.{} must be non-naive datetime.datetime, but it is naive!".
                        format(type(obj).__name__, field_name))


def dataclass_field_listof_check(obj: Any, field_name: str, list_item_type: type) -> None:
    field_val = getattr(obj, field_name)
    if not isinstance(field_val, list):
        raise TypeError("{}.{} must be {}, not {}".
                        format(type(obj).__name__, field_name, list.__name__, type(field_val).__name__))
    for item in field_val:
        if not isinstance(item, list_item_type):
            raise TypeError("{}.{} items must be {}, not {}".
                            format(type(obj).__name__, field_name, list_item_type.__name__, type(item).__name__))


def dataclass_field_optional_check(obj: Any, field_name: str, optional_of_type: type) -> None:
    field_val = getattr(obj, field_name)
    if field_val is None:
        return
    if not isinstance(field_val, optional_of_type):
        raise TypeError("{}.{} must be Optional[{}], not Optional[{}]".
                        format(type(obj).__name__, field_name, optional_of_type.__name__, type(field_val).__name__))


def dataclass_field_optional_datetime_check(obj: Any, field_name: str) -> None:
    field_val = getattr(obj, field_name)
    if field_val is None:
        return
    if not isinstance(field_val, datetime.datetime):
        raise TypeError("{}.{} must be Optional[datetime.datetime], not Optional[{}]".
                        format(type(obj).__name__, field_name, type(field_val).__name__))
    if field_val.tzinfo is None:
        raise TypeError("{}.{} must be None or a non-naive datetime.datetime, but it is naive!".
                        format(type(obj).__name__, field_name))


def dataclass_field_optional_union_check(obj: Any, field_name: str, *args: type) -> None:
    field_val = getattr(obj, field_name)
    if field_val is None:
        return
    for optional_of_type in args:
        if isinstance(field_val, optional_of_type):
            return
    raise TypeError("{}.{} must be Optional[{}], not Optional[{}]".
                    format(type(obj).__name__,
                           field_name,
                           [optional_of_type.__name__ for optional_of_type in args],
                           type(field_val).__name__)
                    )


def dataclass_field_union_check(obj: Any, field_name: str, *args: type) -> None:
    field_val = getattr(obj, field_name)
    for acceptable_type in args:
        if isinstance(field_val, acceptable_type):
            return
    raise TypeError("{}.{} must be one of {}, not {}".
                    format(type(obj).__name__,
                           field_name,
                           [acceptable_type.__name__ for acceptable_type in args],
                           type(field_val).__name__)
                    )


def _check_list_type_elements_helper(obj, field_name: str, list_val: Any, list_element_type):
    for i in list(list_val):
        if not isinstance(i, list_element_type):
            raise TypeError("{}.{} must contain list of {}, but the list has a {}".format(
                type(obj).__name__, field_name,
                _field_type_name(list_element_type), type(i).__name__))
        if issubclass(list_element_type, datetime.datetime) and i.tzinfo is None:
            raise TypeError("{}.{} must contain list of non-naive {}, but the list has a naive {}".format(
                type(obj).__name__, field_name,
                _field_type_name(list_element_type), type(i).__name__))


def _check_dict_type_value_helper(obj, field_name: str, dict_val: Any, dict_value_type):
    for i in dict_val.values():
        if not isinstance(i, dict_value_type):
            raise TypeError("{}.{} must contain dict with values of {}, but the value is a {}".format(
                type(obj).__name__, field_name,
                _field_type_name(dict_value_type), type(i).__name__))
        if issubclass(dict_value_type, datetime.datetime) and i.tzinfo is None:
            raise TypeError("{}.{} must contain list of non-naive {}, but the list has a naive {}".format(
                type(obj).__name__, field_name,
                _field_type_name(dict_value_type), type(i).__name__))


def _dataclass_field_auto_type_check(obj, field_name, field_val, field_type):
    # Note: this gets angry about naive datetime, and won't allow them

    # Union also cover Optional[X], which is Union[X, NoneType]
    if hasattr(field_type, '__origin__') and field_type.__origin__ is Union:
        for allowed_type in field_type.__args__:
            if hasattr(allowed_type, '__origin__') and allowed_type.__origin__ is list:
                assert hasattr(allowed_type, '__args__')
                if isinstance(field_val, list):
                    list_element_type = allowed_type.__args__[0]
                    _check_list_type_elements_helper(obj, field_name, field_val, list_element_type)
                    return
            else:
                if isinstance(field_val, allowed_type):
                    must_reraise = False
                    try:
                        if issubclass(allowed_type, datetime.datetime) and field_val.tzinfo is None:
                            must_reraise = True
                            raise TypeError("{}.{} must be {}, and it is a {}, but a naive one".format(
                                type(obj).__name__, field_name,
                                _field_type_name(allowed_type), type(field_val).__name__))
                        return
                    except:
                        # it's not a datetime.datetime, so we can ignore that issubclass doesn't work at this point
                        # (if allowed_type=typing.Dict then issubclass doesn't work)
                        if must_reraise:
                            raise
                        else:
                            return
                        # raise TypeError("{}.{} must be {}, but it is {} and "
                        #                 "something went wrong because of {!r}".format(
                        #     type(obj).__name__, field_name,
                        #     _field_type_name(allowed_type), type(field_val).__name__,
                        #     allowed_type))
        raise TypeError("{}.{} must be one of {}, not {} ({!r} {!r})".format(
            type(obj).__name__, field_name,
            [_field_type_name(t) for t in field_type.__args__], type(field_val).__name__,
            field_type.__args__, type(field_val)))

    if hasattr(field_type, '__origin__') and field_type.__origin__ is list:
        list_element_type = field_type.__args__[0]
        if not isinstance(field_val, list):
            raise TypeError("{}.{} must be list, not {}".format(
                type(obj).__name__, field_name, type(field_val).__name__))
        _check_list_type_elements_helper(obj, field_name, field_val, list_element_type)
        return

    if hasattr(field_type, '__origin__') and field_type.__origin__ is dict:
        if not field_type.__args__:
            field_type = dict
            # no return, just check if it's a generic dict
        else:
            dict_key_type = field_type.__args__[0]
            dict_value_type = field_type.__args__[1]
            if dict_key_type is not str:
                raise NotImplementedError("deep Dict type check with non-str keys is not implemented")
            if not isinstance(field_val, dict):
                raise TypeError("{}.{} must be dict, not {}".format(
                    type(obj).__name__, field_name, type(field_val).__name__))
            _check_dict_type_value_helper(obj, field_name, field_val, dict_value_type)
            return

    if not isinstance(field_val, field_type):
        raise TypeError("{}.{} must be {}, not {}".format(
            type(obj).__name__, field_name,
            _field_type_name(field_type), type(field_val).__name__))

    if issubclass(field_type, datetime.datetime) and field_val.tzinfo is None:
        raise TypeError("{}.{} must be non-nave {}, not is it a naive {}".format(
            type(obj).__name__, field_name,
            _field_type_name(field_type), type(field_val).__name__))


def dataclass_field_auto_type_check(obj, field_name):
    field = next(field for field in dataclasses.fields(obj) if field.name == field_name)
    field_val = getattr(obj, field_name)
    _dataclass_field_auto_type_check(obj, field_name, field_val, field.type)


def dataclass_fields_type_check(
        obj: Any,
        field_names_included: Optional[List[str]] = None,
        field_names_excluded: Optional[List[str]] = None) -> None:
    for field in dataclasses.fields(obj):
        if (not field_names_included or field.name in field_names_included) and \
                not (field_names_excluded and field.name in field_names_excluded):
            field_val = getattr(obj, field.name)
            _dataclass_field_auto_type_check(obj, field.name, field_val, field.type)
            # if not isinstance(field_val, field.type):
            #     raise TypeError("{}.{} must be {}, not {}".format(type(obj).__name__, field.name, _field_type_name(field.type), type(field_val).__name__))


def _add_auto_type_check_to_class(cls,
                                  field_names_included: Optional[List[str]] = None,
                                  field_names_excluded: Optional[List[str]] = None):

    has_post_init = hasattr(cls, '__post_init__')

    if has_post_init:
        # print('adding auto type check to {!r} (adding to __post_init__)'.format(cls))
        old_post_init = cls.__post_init__

        def updated_post_init(self, *args, **kwargs):
            old_post_init(self, *args, **kwargs)
            dataclass_fields_type_check(self, field_names_included, field_names_excluded)
        cls.__post_init__ = updated_post_init
    else:
        # print('adding auto type check to {!r} (new __post_init__)'.format(cls))

        def new_post_init(self):
            dataclass_fields_type_check(self, field_names_included, field_names_excluded)
        cls.__post_init__ = new_post_init

    return cls


def _add_copy_method_to_class(cls):
    def make_copy(self):
        return copy.deepcopy(self)

    cls.make_copy = make_copy
    return cls


def dataclass_auto_type_check(_cls=None, *,
                              field_names_included: Optional[List[str]] = None,
                              field_names_excluded: Optional[List[str]] = None):
    """
        This assumes @dataclass !  (but not @dataclass_json or @dataclass_dict_convert)
        It must be placed both BELOW @dataclass/@dataclass_dict_convert !
                (otherwise, it won't work if you have no __post_init__ yet!!!)

    this has complex logic to allow decorating with both:
          @dataclass_auto_type_check
          @dataclass_auto_type_check(field_names_excluded=['aFieldName'])

    example:
          @dataclass
          @dataclass_auto_type_check(field_names_excluded=['field_a'])
          class X:
             field_a: int
             ...
    """

    def wrap(cls):
        return _add_auto_type_check_to_class(cls, field_names_included, field_names_excluded)

    if _cls is None:
        return wrap
    return wrap(_cls)


def dataclass_copy_method(_cls=None):
    """
    This assumes @dataclass
    It works both below and above @dataclass_json/@dataclass_dict_convert and @dataclass

    this has complex logic to allow decorating with both:
      @dataclass_copy_method
      @dataclass_copy_method()

    example:
      @dataclass_copy_method
      @dataclass_dict_convert
      @dataclass
      class X:
         field_a: int
         ...
    """

    def wrap(cls):
        return _add_copy_method_to_class(cls)

    if _cls is None:
        return wrap
    return wrap(_cls)


def _add_dataclass_multiline_repr_method_to_class(cls):
    # recursive_repr function is based on "recursive_repr" function in reprlib and dataclasses _repr_fn
    # This version doesn't use helper functions

    repr_running = set()

    def dataclasses_multiline_repr_helper(self):
        key = _thread.get_ident()
        if key in repr_running:
            return '...'
        repr_running.add(key)
        try:
            all_fields = dataclasses.fields(self)
            field_repr_list = []
            for field in all_fields:
                field_val = getattr(self, field.name)
                field_repr = f"   {field.name}={field_val!r}"
                field_repr = field_repr.replace('\n', '\n   ')
                field_repr_list.append(field_repr)
            fields_repr = '\n' + ',\n'.join(field_repr_list)
            result = f"{self.__class__.__qualname__}({fields_repr})"
        finally:
            repr_running.discard(key)
        return result

    cls.__repr__ = dataclasses_multiline_repr_helper
    return cls


def dataclass_multiline_repr(_cls=None):
    """
    Make repr use newlines between fields, for much more readable output.

    This assumes @dataclass
    It works both below and above @dataclass, but if above @dataclass, you need: @dataclass(repr=False)

    this has complex logic to allow decorating with both:
      @dataclass_multiline_repr
      @dataclass_multiline_repr()

    example:
      @dataclass_multiline_repr
      @dataclass(repr=False)
      class X:
         field_a: int
         ...

    or same:
      @dataclass
      @dataclass_multiline_repr
      class X:
         field_a: int
         ...

    """
    def wrap(cls):
        return _add_dataclass_multiline_repr_method_to_class(cls)

    if _cls is None:
        return wrap
    return wrap(_cls)



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="dataclass-dict-convert",
    version="1.0.1",

    description="Convert between Dataclasses and dict/json",
    long_description=long_description,
    long_description_content_type="text/markdown",

    url="https://gitlab.ilabt.imec.be/wvdemeer/dataclass-dict-convert",

    author="Wim Van de Meerssche",
    author_email="wim.vandemeerssche@ugent.be",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],

    packages=["dataclass_dict_convert"],
    include_package_data=True,

    install_requires=[
        "python-dateutil",
        "stringcase",  # used in tests, AND typically by users of the lib. But not by the lib itself.
    ],
    python_requires='>=3.7',   # python >= 3.7 because python dataclasses are required

    # Tests with py.test:  run with: python setup.py pytest
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    zip_safe=False,
)
